"""polka: Implementation of polarimetric models"""

import lmfit
import numpy as np
from scipy.optimize import root as scipy_root

MODELS = ["LinExp", "Trigo"]
# MODELS = ["LinExp"]


class LinExp:
    """Linear-Exponential model from Muinonen+2002"""

    def __init__(self, a=np.nan, b=np.nan, c=np.nan):
        self.NAME = "LinExp"
        self.PARAMS = ("a", "b", "c")

    def eval(self, phase, a=None, b=None, c=None):
        """Evaluate LinExp model for given phase and parameters."""
        aa = a if a is not None else getattr(self, "a")
        bb = b if b is not None else getattr(self, "b")
        cc = c if c is not None else getattr(self, "c")
        return aa * (np.exp(-phase / bb) - 1) + cc * phase

    def fit(self, pc, weights=None):
        """Fit a phase curve using the Linear-Exponential model."""

        # LMFit model definition
        model = lmfit.Model(self.eval)

        # LMFit parameters
        params = lmfit.Parameters()
        params.add("a", value=15, min=0, max=30)
        params.add("b", value=10, min=0, max=20)
        params.add("c", value=0.1, min=-10, max=10)

        if weights is not None:
            weights_band = weights
        else:
            weights_band = np.ones(pc.phase.shape)

        # And fit
        result = model.fit(
            pc.pol,
            params,
            phase=pc.phase,
            weights=weights,
            method="least_squares",
            fit_kws={"loss": "soft_l1"},
        )

        # Add fit parameters to the model instance
        for param in self.PARAMS:
            setattr(self, param, result.params[param].value)
            setattr(self, f"{param}_err", result.params[param].stderr)

        # Add derived parameters
        self.alpha_min = -self.b * np.log(self.b * self.c / self.a)
        self.pol_min = self.eval( self.alpha_min, self.a, self.b, self.c)

        sol = scipy_root(
            self.eval,
            25.0,
            args=(self.a, self.b, self.c),
        )
        self.alpha_inv = sol["x"][0] if sol["success"] else np.nan

        self.slope = self.c - (self.a/self.b) * np.exp(-self.alpha_inv / self.b)

        pc.fitted_models.add("LinExp")





class Trigo:
    """Trigonometric model from Lumme & Muinonen 1993"""

    def __init__(self, a=np.nan, b=np.nan, c=np.nan):
        self.NAME = "Trigo"
        self.PARAMS = ("b", "c1", "c2", "alpha_inv")

    def eval(self, phase, b=None, c1=None, c2=None, alpha_inv=None):
        """Evaluate LinExp model for given phase and parameters."""
        bb = b if b is not None else getattr(self, "b")
        cc1 = c1 if c1 is not None else getattr(self, "c1")
        cc2 = c2 if c2 is not None else getattr(self, "c2")
        alpha_inv = alpha_inv if alpha_inv is not None else getattr(self, "alpha_inv")
        ph = np.radians(phase)
        return bb * (np.sin(ph) ** cc1) * (np.cos(ph / 2) ** cc2) * np.sin(ph - np.radians(alpha_inv))

    def fit(self, pc, weights=None):
        """Fit a phase curve using the Trigonometric model."""

        # LMFit model definition
        model = lmfit.Model(self.eval)

        # LMFit parameters
        params = lmfit.Parameters()
        params.add("b", value=1, min=-20, max=20)
        params.add("c1", value=1, min=-20, max=20)
        params.add("c2", value=1, min=-20, max=20)
        params.add("alpha_inv", value=15, min=0, max=40)

        if weights is not None:
            weights_band = weights
        else:
            weights_band = np.ones(pc.phase.shape)

        # And fit
        result = model.fit(
            pc.pol,
            params,
            phase=pc.phase,
            weights=weights,
            method="least_squares",
            fit_kws={"loss": "soft_l1"},
        )

        # Add fit parameters to the model instance
        for param in self.PARAMS:
            setattr(self, param, result.params[param].value)
            setattr(self, f"{param}_err", result.params[param].stderr)

        # Add derived parameters
        # self.pol_min = self.eval( self.alpha_min, self.a, self.b, self.c)

        # sol = scipy_root(
        #     self.eval,
        #     25.0,
        #     args=(self.a, self.b, self.c),
        # )
        # self.alpha_inv = sol["x"][0] if sol["success"] else np.nan

        # self.slope = self.c - (self.a/self.b) * np.exp(-self.alpha_inv / self.b)

        pc.fitted_models.add("Trigo")


import numpy as np
import matplotlib.pyplot as plt

import polka
import rocks


def plot_pc(
    pc, models, label_sources=False, show_parameters=False, black=False, save=None
):
    """Plot polrimetric phase curve and model fits."""

    # Define figure
    if black:
        plt.style.use("dark_background")
    else:
        plt.style.use("default")
    fig, ax = plt.subplots()

    # Global phase range
    xmax = pc.phase.max() + 3
    phase_eval = np.linspace(0, xmax, 100)

    # Observations
    if label_sources and pc.src is not None:

        for s in np.unique(pc.src):
            cond = np.where(pc.src == s)
            ax.errorbar(
                pc.phase[cond],
                pc.pol[cond],
                yerr=pc.pol_err[cond],
                ls="",
                marker="o",
                label=s,
            )
    else:
        ax.errorbar(
            pc.phase,
            pc.pol,
            yerr=pc.pol_err,
            ls="",
            marker="o",
            label=f"Observations",
        )

    # Models
    for i, model in enumerate(models):

        # Switch to actual model instance
        model = getattr(pc, model)
        params = model.PARAMS

        # Plot model
        pol_eval = model.eval(phase_eval)
        ax.plot(phase_eval, pol_eval, color=f"C{i+1}", ls="-", label=model.NAME)

        # Derived parameters
        if show_parameters:
            ax.axhline(model.pol_min, ls="--", color=f"C{i+1}", lw=0.5)
            ax.axvline(model.alpha_min, ls="--", color=f"C{i+1}", lw=0.5)
            ax.axvline(model.alpha_inv, ls="dotted", color=f"C{i+1}", lw=0.5)

    # Axes
    ax.set(
        xlabel="Phase angle (deg.)",
        ylabel="Linear polarisation (%)",
        xlim=(0, xmax),
    )
    ax.axhline(0, ls="-", color="gray", lw=0.5)

    # Legend
    ax.legend(
        title=(
            f"({pc.target.number}) {pc.target.name}"
            if isinstance(pc.target, rocks.Rock)
            else None
        )
    )

    # Export
    if save is None:
        plt.show()
    else:
        fig.tight_layout()
        fig.savefig(save)
        print(f"Saved figure under {save}")

    return fig, ax



import numpy as np
import polka
import rocks


class PhaseCurve:
    """Polarimetric phase curve of the asteroid."""

    def __init__(
        self,
        phase=None,
        pol=None,
        pol_err=None,
        src=None,
        epoch=None,
        target=None,
    ):
        """Create a polarimetric phase curve.

        Parameters
        ----------
        phase : list or np.array
            Phase angle (in degrees).
        pol : list or np.array
            Degree of linear polarisation (in percent).
        pol_err : list or np.array
            Error on the degree of linear polarisation (in percent).
        src : list
            Source of the observation.
        epoch : list or np.array
            Epoch of observation (in JD).
        target : int or str
            The target asteroid.
        """

        # Observations
        self.phase = np.array(phase)
        self.pol = np.array(pol)
        self.pol_err = (
            np.array(pol_err) if pol_err is not None else np.zeros(self.pol.shape)
        )
        self.src = np.array(src)
        self.epoch = np.array(epoch) if epoch is not None else epoch

        # Metadata
        self.target = target
        if target is not None:
            self.target = rocks.Rock(target)

        # Modeling
        self.fitted_models = set()  # keep track of models fit to data

    def fit(self, models=None, weights=None):
        """Fit the polarimetric phase curve with the different models."""

        if models is None:
            models = polka.models.MODELS

        for model in models:
            if model not in polka.models.MODELS:
                raise ValueError(
                    f"Unknown model '{model}'. Expected one of {polka.models.MODELS}"
                )

            # Add polarimetric model instance to PhaseCurve
            setattr(self, model, getattr(polka.models, model)())
            getattr(self, model).fit(self, weights=weights)

    def get_ephems(self):
        """Query ephemerides of target at time of observations.

        Note
        ----
        Sets the 'phase' attribute. Requires internet connection.
        """
        print("Querying ephemerides via LTE Miriade..")
        ephem = polka.miriade.query(self.target.name, self.epoch)
        self.phase = ephem["Phase"].to_numpy()

    def plot(
        self,
        models=None,
        label_sources=False,
        show_parameters=False,
        black=False,
        save=None,
    ):
        """Plot polarimetric phase curve and model fits.

        Parameters
        ----------
        models : list of str
            Name of models to plot. By default, all fitted models are plotted.
        """
        if models is None:
            models = sorted(self.fitted_models)

        f, a = polka.plotting.plot_pc(
            self, models, label_sources, show_parameters, black, save
        )

        return f, a



from astropy.time import Time
import pandas as pd
import requests


def query(name, epochs):
    """Gets asteroid ephemerides from LTE Miriade.

    Parameters
    ----------
    name : str
        Name or designation of asteroid.
    epochs : list
        List of observation epochs in JD or ISO formats.

    Returns
    -------
    pd.DataFrame - Ephemerides for the requested epochs
                   False - If query failed somehow
    """

    # Pass sorted list of epochs to speed up query
    # Convert epochs to JD
    epochs = [ e if isinstance(e, float) else float(Time(e).jd) for e in epochs ]
    files = {"epochs": ("epochs", "\n".join(["%.6f" % epoch for epoch in epochs]))}

    # VOSSP/Miriade parameters
    url = "https://ssp.imcce.fr/webservices/miriade/api/ephemcc.php"

    params = {
        "-name": f"a:{name}",
        "-mime": "json",
        "-tcoor": "1",
        "-output": "--jd",
        "-tscale": "UTC",
        "-observer": "500",
    }

    # Execute query
    try:
        r = requests.post(url, params=params, files=files, timeout=50)
    except requests.exceptions.ReadTimeout:
        return False
    j = r.json()

    # Read JSON response
    try:
        ephem = pd.DataFrame.from_dict(j["data"])
    except KeyError:
        return False

    return ephem
